import re
import io
import os
import time
import nltk
import config # config that has the file paths in it
import warnings
import multiprocessing
from nltk.corpus import stopwords
warnings.filterwarnings(action='ignore', category=UserWarning, module='gensim')
import gensim

nltk_path = config.NLTK_PATH # path to NLTK tokenizer files
sent_path = config.SENT_PATH # path to SentiWS text files
sent_pos = [] # positive words from SentiWS
sent_neg = [] # negative words from SentiWS
stop_words = stopwords.words('german') # initialize stopwords


# initialize word lists for SentiWS
# sentiment: either 'positive' or 'negative'
# word_list: either sent_pos or sent_neg
def init_sent(sentiment, word_list):
    # get words from file
    with open('{}{}.txt'.format(sent_path, sentiment), encoding='utf-8') as file:
        sent_words_all = file.readlines()

    # iterate over the lines and get words + value from it and append it to the respective list
    for line in sent_words_all:
        word_info = line.split()
        base_word = word_info[0].split('|')[0]  # splits first word into 'word','|','XX' and then accesses word
        base_word = base_word.replace('ß', 'ss').replace('ä', 'ae').replace('ö', 'oe').replace('ü', 'ue')
        sent_value = float(word_info[1])
        word_list.append([base_word, sent_value])
        if len(word_info) > 2: # check if there are word forms other than the base form
            for word in word_info[2].split(','):
                word = word.replace('ß', 'ss').replace('ä', 'ae').replace('ö', 'oe').replace('ü', 'ue')
                word_list.append([word, sent_value])


# creates the sentiment score for two similarity values
# pos, neg: positive similarity value and negative similarity value of the word
# score is based on the difference of pos and neg
def sent_score(pos, neg):
    if pos >= neg:
        diff = pos - neg
        if diff <= 0.05:
            score = 0.2
        elif diff <= 0.1:
            score = 0.4
        elif diff <= 0.15:
            score = 0.6
        elif diff <= 0.2:
            score = 0.8
        else:
            score = 1

    else:
        diff = neg - pos
        if diff <= 0.05:
            score = -0.2
        elif diff <= 0.1:
            score = -0.4
        elif diff <= 0.15:
            score = -0.6
        elif diff <= 0.2:
            score = -0.8
        else:
            score = -1
    return score


# initialized with site name as parameter
# creates a word2vec object that can fetch articles from text files, make them into word lists and create a w2v model from them
# the created word2vec model can be stored and loaded
class word2vec(object):
    def __init__(self, site):
        self.site = site
        self.file_path = config.INPUT_DIR+site+'.txt'
        self.text = ''
        self.sentences = []
        if os.path.isfile('trained/{}_model.w2v'.format(self.site)):
            self.model = self.load_model()
        else:
            self.model = None

    # gets text from all the articles of the respective news outlet from .txt file
    def get_text_from_files(self):
        now = time.time()
        print('Fetching text from {}.txt'.format(self.site))
        with open('{}'.format(self.file_path), 'r', encoding='utf-8') as text_file:
            self.text += text_file.read()
            print('Finished fetching text from {}.txt in {}s'.format(self.site, round(time.time()-now, 2)))
        print('The text consists of {:,} letters.'.format(len(self.text)))
        print('-----------------------')

    # splits corpus into sentences
    def create_sentence_list(self):
        print('Splitting the corpus into sentences..')
        now = time.time()
        tokenizer = nltk.data.load('file://' + nltk_path)
        raw_sentences = tokenizer.tokenize(self.text)
        for raw_sentence in raw_sentences:
            if len(raw_sentence) > 0:
                self.sentences.append(self.sentence_to_wordlist(raw_sentence))
        print('Finished creating sentences in {}s'.format(round(time.time() - now, 2)))
        word_amount = sum([len(sentence) for sentence in self.sentences])
        print('The text consists of {0:,} words.'.format(word_amount))
        print('-----------------------')

    # filter unwanted characters and stopwords
    # takes a sentence in the form of a string as input
    # returns a word list
    @staticmethod
    def sentence_to_wordlist(sentence):
        clean_sentence = re.sub('[^a-zA-ZäöüÄÖÜß]', ' ', sentence)
        word_list = [w for w in clean_sentence.split() if not w in stop_words]
        return word_list

    # saves word2vec model
    def store_model(self):
        self.model.save('trained/{}_model.w2v'.format(self.site))

    # loads word2vec model from file
    def load_model(self):
        return gensim.models.Word2Vec.load('trained/{}_model.w2v'.format(self.site))

    # creates word vectors for the specified sentences and stores the model
    # min_count: amount of times a word has to appear to be added to the vocabulary
    # size: dimensionality of the word vectors
    # workers: number of threads
    # sample: random number for downsampling of the high frequency words
    def create_model(self):
        print('Creating word vectors..')
        now = time.time()
        self.model = gensim.models.Word2Vec(sentences=self.sentences, min_count=25, size=500, workers=multiprocessing.cpu_count(), sample=1e-3)
        print('Finished creating word vectors in {}s'.format(round(time.time() - now, 2)))
        print('Storing model..')
        self.store_model()
        print('-----------------------\n')


if __name__ == '__main__':
    # download nltk data for tokenizing process
    nltk.download('punkt')

    # initialize sent words
    init_sent('positive', sent_pos)
    init_sent('negative', sent_neg)

    for site in ['faz', 'focus', 'spiegel', 'sueddeutsche', 'tagesschau', 'tonline', 'welt', 'zdf', 'zeit']:
        # create word2vec object and train model if there is no pretrained model saved
        w2v = word2vec(site)
        if w2v.model is None:
            w2v.get_text_from_files()
            w2v.create_sentence_list()
            w2v.create_model()

        # load test articles from file
        with open('test_articles/{}.txt'.format(site), encoding='utf-8') as test_file:
            test_articles = test_file.readlines()

        # start of sentiment analysis of the articles from the respective site
        print('Analyzing {} articles..'.format(site))

        # amount of test articles that you want to analyze
        articles_to_analyze = 50
        article_amount = 0
        duplicates = []

        # analyze each article until wanted amount is reached
        for test_article in test_articles:
            # check if limit is reached and article has not already been analyzed
            if article_amount < articles_to_analyze and test_article not in duplicates:
                duplicates.append(test_article)

                # create word list from test article
                words = []
                article_words = w2v.sentence_to_wordlist(test_article)

                # add all words from test article that are included in the model's vocabulary to a list
                for test_word in article_words:
                    if test_word in w2v.model.wv.vocab:
                        words.append(test_word)

                # create scores for article
                article_score = 0

                # only analyze article if it has more than 10 words
                if len(words) > 10:
                    article_amount += 1

                    # calculate a sentiment score for each word in the article's word list iteratively
                    for word in words:
                        pos_amount = 0  # amount of positive words from SentiWS that are in the model's vocabulary
                        neg_amount = 0  # amount of negative words from SentiWS that are in the model's vocabulary
                        pos_pos = 0 # amount of positive words from SentiWS that have a positive similarity to word
                        pos_neg = 0 # amount of positive words from SentiWS that have a negative similarity to word
                        neg_pos = 0 # amount of negative words from SentiWS that have a positive similarity to word
                        neg_neg = 0 # amount of negative words from SentiWS that have a negative similarity to word

                        # check how many positive words from SentiWS have a positive/negative similarity
                        for pos_word in sent_pos:
                            if pos_word[0] in w2v.model.wv.vocab:
                                pos_amount += 1
                                # similarity value between -100 and 100
                                similarity = w2v.model.wv.similarity(word, pos_word[0]) * 100
                                # increment based on similarity
                                if similarity < 0:
                                    pos_neg += 1
                                else:
                                    pos_pos += 1

                        # check how many negative words from SentiWS have a positive/negative similarity
                        for neg_word in sent_neg:
                            if neg_word[0] in w2v.model.wv.vocab:
                                neg_amount += 1
                                # similarity value between -100 and 100
                                similarity = w2v.model.wv.similarity(word, neg_word[0]) * 100
                                # increment based on similarity
                                if similarity < 0:
                                    neg_neg += 1
                                else:
                                    neg_pos += 1

                        # calculate percentage amount of positive similarities for positive/negative SentiWS words [0:1]
                        pos_sim = pos_pos / pos_amount
                        neg_sim = neg_pos / neg_amount

                        # if positive similarity is over 90% for either of the positive or negative words
                        # or the difference between the positive similarity of the positive and negative words is >= 0.3
                        # the word is seen as positive/negative and a sentiment score is determined
                        # if neither of these conditions is met, the word is seen as neutral and gets the score 0
                        if pos_sim > 0.9 or neg_sim > 0.9 or abs(pos_sim - neg_sim) >= 0.3:
                            word_score = sent_score(pos_sim, neg_sim)
                        else:
                            word_score = 0

                        # add word score to current article score
                        article_score += word_score

                    # determine sentiment (word) by sentiment score
                    # 0 -> neutral, > 0 -> positive, < 0 -> negative
                    if article_score == 0:
                        sent = 'neutral'
                    elif article_score > 0:
                        sent = 'positive'
                    else:
                        sent = 'negative'

                # write article #, article and sentiment score to file
                with io.open('output/{}.txt'.format(site), 'a', encoding='utf-8') as output_file:
                    output_file.write('#{}\n'.format(article_amount))
                    output_file.write(test_article)
                    output_file.write('Article Sentiment Score: {} ({})\n'.format(round(article_score, 2), sent, len(test_article.split())))

                # write just the sentiment score to a separate file (for diagram creation etc)
                # remove '.replace('.', ',')' if you need '.' as number separator instead of ',' for diagrams
                with io.open('output/scores_{}.txt'.format(site), 'a', encoding='utf-8') as output_file:
                    output_file.write(str(round(article_score, 2)).replace('.', ',') + '\n')
